// Virtual COM port implementation
// Originally published by W.S. at Nuvoton, improved by members of microcontroller.net
#ifndef __USB_H__
#define __USB_H__

#include <stdint.h>
#include <stdbool.h>

// ==========================================================================================================
// Description of the USB driver
// ==========================================================================================================
/*
            Scope of usb.c
       ___________________________
      /                           \
 ______         ________         ___________             ___________         ________         ______
|      |       |        |       |           |           |           |       |        |       |      |
|      |--->---| Tx buf |--->---|           |           |           |--->---| Rx buf |--->---|      |
| Your |       |________|       | STM32 USB |           | PC USB    |       |________|       | PC   |
| app  |        ________        | interface |===========| interface |        ________        | app  |
|      |       |        |       |           |           |           |       |        |       |      |
|      |---<---| Rx buf |---<---|           |           |           |---<---| Tx buf |---<---|      |
|______|       |________|       |___________|           |___________|       |________|       |______|
   |                                 |                         |                                 |
   |                                 |                         |                                 |
   +---------------------------->----+                         +----<----------------------------+
         Control commands                                                Control commands


The USB communication is driven by the PC. The PC tells when it wants to receive or send data and how
much (up to 64 bytes per package). In addition the PC sends a SOF package every 1 ms to check if the
connection is alive.

When the SOF packages are missing, then the USB interface goes into suspend mode. Then the app may read
remaining characters from the Rx buffer but sending is blocked (except function UsbSendChar_noWait(),
which can fill the Tx buffer anyway).

Whenever the µC receives a package from the PC, an interrupt occurs. So the whole communication is driven
by interrupts. There is no need for a background thread. The ISR must respond every package, or the PC
will drop the connection (after some few repetitions). You have to avoid blocking interrupts for more
than one 1ms.

The application interface provides many functions which can wait until a timeout of some milliseconds.
Please note that this timeout depends on counting the regular SOF packages from PC. So the actual time
may be a little longer than requested, specially when the connection drops.

When the connection drops, the buffers are not cleared. You can re-establish the connection, then the
data transfer continues without loosing a single character.

There is no way for the STM32 to find out if an application on the PC side has opened the virtual COM port.
If the virtual COM port is not opened, the Tx buffer will fill up quickly. But you can check the USB
connection with function UsbActive().

When the Rx buffer of the STM32 is full, then the PC cannot send more data to the STM32. In this case,
the Tx buffer on the PC side will fill up. It's up to the PC application what happens then.

For the other direction where the STM32 send data to the PC, a timeout may occur when the communication
gets stuck. The API reports back how many characters have been written into the buffer. Your application
must decide what to do in that case (discard the remaining characters or try to repeat).

See configurable settings in usb.c
*/

// ==========================================================================================================
// Application interface
// ==========================================================================================================

// Start and initialize USB/HID
/* VID, PID: these are used to identify the device.
 * vendor_string, product_string, serial_number: these are the string descriptors that will show up on the PC.
 *      Values can be 0, but memory location must remain unaltered (i.e. buffer content must be const)
 *      Strings must be Null-terminated and no longer than 30 bytes.
 * polling_rate: the frequency at which the host queries the device. Value in ms, must not be 0.
 * void cbDataRx(void): Callback function if data has been received from host.
 *      The application MUST then call UsbFetchData() to retrieve the data and re-enable reception.
 *      Value can be 0, then data reception from the host (OUT transmission) is disabled.
 */
typedef void (*usb_cb_t)(void);

void UsbSetup(const uint16_t VID, const uint16_t PID, const uint8_t *vendor_string, const uint8_t *product_string, const uint8_t *serial_number, const uint16_t polling_rate, usb_cb_t cbDataRx);

// Get recently received data (and turn reception back on)
bool UsbFetchData(uint8_t *buffer);

// Send data back to host. Exactly REPORT_COUNT bytes will be sent
bool UsbSendData(uint8_t *data);

#endif
